"""Viper tasks
--------------
A viper task can be defined using `viper.Task(name, command_factory)`
where name is the name of the task which should help easily recognise the
purpose of the task and `command_factory` generates the command (see `viper.demo.commands`).

We can also pass optional arguments `timeout`, `retry`, `pre_run` and `post_run`.

`timeout` is applied to the duration of execution of the command generated
by the `command_factory`. If `retry` is more than 0, the task will re-run itself.

`pre_run` and `post_run` are callbacks. `pre_run` is run just before the command
if about to execute. `post_run` is run after the result of the task is saved in the DB
but before retry in case the command fails (see `viper.demo.callbacks`).
"""

from viper import Task
from viper.demo.callbacks import log_command
from viper.demo.callbacks import log_status
from viper.demo.commands import df_command
from viper.demo.commands import ping_command
from viper.demo.commands import remote_execute_command


def ping() -> Task:
    """Ping the host one time"""

    return Task(
        "Ping",
        ping_command,
        timeout=5,
        retry=1,
        pre_run=log_command,
        post_run=log_status,
    )


def remote_execute() -> Task:
    """Execute a command on the host"""

    return Task(
        "Remote execute",
        remote_execute_command,
        timeout=300,
        retry=0,
        pre_run=log_command,
        post_run=log_status,
    )


def disk_usage() -> Task:
    """Get the local disk usage of the host"""

    return Task(
        name="Disk Usage",
        command_factory=df_command,
        timeout=5,
        retry=1,
        pre_run=log_command,
        post_run=log_status,
    )



"""Project specific viper commands."""

from argparse import FileType
from argparse import Namespace
from time import ctime
from viper import Host
from viper import Hosts
from viper import Result
from viper import Results
from viper import Runner
from viper import Task
from viper.project import Arg
from viper.project import Project

import csv
import json
import sys
import typing as t

myproj = Project(prefix="myproj")


def log_command_callback(runner: Runner) -> None:
    """Log the command before run."""
    command = runner.task.command_factory(runner.host, *runner.args)
    print(runner.host.ip, command, file=sys.stderr, sep=": ")


def log_status_callback(result: Result) -> None:
    """Log the status after task run."""

    if result.ok():
        print(result.host.ip, "OK", file=sys.stderr, sep=": ")
        return

    print(result.host.ip, "FAILED", file=sys.stderr, sep=": ")


@myproj.hostgroup(
    args=[
        Arg("-f", "--file", type=FileType("r"), default="hosts.json"),
        Arg("-i", "--identity_file", default="/root/.ssh/id_rsa.pub"),
    ]
)
def allhosts(args) -> Hosts:
    """Get all the myproj hosts."""

    data = json.load(args.file)

    return Hosts.from_items(
        *(
            Host(
                ip=d["ip"],
                hostname=d["name"],
                login_name="root",
                identity_file=args.identity_file,
                meta=tuple(d.items()),
            )
            for d in data
        )
    )


@myproj.filter(objtype=Hosts, args=[Arg("key"), Arg("val")])
def hosts_by(host: Host, args: Namespace) -> bool:
    """Filter hosts by key and metadata"""

    return str(dict(host.meta)[args.key]) == args.val


@myproj.filter(objtype=Results, args=[Arg("key"), Arg("val")])
def results_by(result: Result, args: Namespace) -> bool:
    """Filter hosts by IP address"""

    return (
        hasattr(result, args.key) and str(getattr(result, args.key)).strip() == args.val
    )


@myproj.handler(fromtype=Hosts, totype=Hosts, args=[Arg("file", type=FileType("w"))])
def hosts2csv(hosts: Hosts, args: Namespace) -> Hosts:
    """Export csv formatted hosts to file"""

    if hosts.count() > 0:
        writer = csv.writer(args.file)
        writer.writerow(list(dict(hosts[0].meta).keys()))
        for host in hosts.all():
            writer.writerow(list(dict(host.meta).values()))
    args.file.flush()
    args.file.close()

    # Printing JSON to terminal to that it can be piped to further commands
    return hosts


@myproj.handler(
    fromtype=Results, totype=Results, args=[Arg("file", type=FileType("w"))]
)
def results2csv(results: Results, args: Namespace) -> Results:
    """Export csv formatted results to file"""

    if results.count() > 0:
        writer = csv.writer(args.file)
        writer.writerow(
            [
                "task",
                "ip",
                "hostname",
                "command",
                "returncode",
                "success",
                "start",
                "end",
                "retry",
                "stdout",
                "stderr",
            ]
        )

        for r in results.all():
            writer.writerow(
                [
                    r.task.name,
                    r.host.ip,
                    r.host.hostname,
                    r.command,
                    r.returncode,
                    r.ok(),
                    ctime(r.start),
                    ctime(r.end),
                    r.retry,
                    r.stdout,
                    r.stderr,
                ]
            )
    args.file.flush()
    args.file.close()

    # Printing JSON to terminal to that it can be piped to further commands
    return results


def remote_exec_command(host: Host, command: str) -> t.Sequence[str]:
    if not host.login_name:
        raise ValueError(f"{host}: 'login_name' is not set")

    if not host.identity_file:
        raise ValueError(f"{host}: 'identity_file' is not set")

    return (
        "ssh",
        "-l",
        host.login_name,
        "-i",
        host.identity_file,
        "-o",
        "StrictHostKeyChecking=no",
        host.ip,
        command,
    )


@myproj.job(
    fromtype=Hosts,
    totype=Results,
    args=[
        Arg("command"),
        Arg("file", type=FileType("w"), help="CSV file path for the result"),
        Arg("--max-workers", default=0, type=int),
    ],
)
def remote_exec(hosts: Hosts, args: Namespace) -> Results:
    """Execute command on hosts remotely."""
    return hosts.run_task(
        Task(
            "Remote execute",
            remote_exec_command,
            timeout=300,
            retry=0,
            pre_run=log_command_callback,
            post_run=log_status_callback,
        ),
        args.command,
        max_workers=args.max_workers,
    ).pipe(lambda results: results2csv(results, args))


def app_version_command(host: Host, app: str) -> t.Sequence[str]:
    return remote_exec_command(host, f"{app} --version")


@myproj.job(
    fromtype=Hosts,
    totype=Results,
    args=[
        Arg("app"),
        Arg("file", type=FileType("w"), help="CSV file path for the result"),
        Arg("--max-workers", default=0, type=int),
    ],
)
def app_version(hosts: Hosts, args: Namespace) -> Results:
    """Get application version."""
    return hosts.run_task(
        Task(
            "Get app version",
            app_version_command,
            timeout=10,
            retry=3,
            pre_run=log_command_callback,
            post_run=log_status_callback,
        ),
        args.app,
        max_workers=args.max_workers,
    ).pipe(lambda results: results2csv(results, args))


def update_via_apt_command(host: Host, app: str) -> t.Sequence[str]:
    return remote_exec_command(host, f"cronly --version && echo updating {app}")


@myproj.job(
    fromtype=Hosts,
    totype=Results,
    args=[
        Arg("app"),
        Arg("version"),
        Arg("file", type=FileType("w"), help="CSV file path for the result"),
        Arg("--max-workers", default=0, type=int),
    ],
)
def update_via_apt(hosts: Hosts, args: Namespace) -> Results:
    """Update application via apt."""
    return (
        hosts.run_task(
            Task(
                "Get app version",
                app_version_command,
                timeout=10,
                retry=3,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.app,
            max_workers=args.max_workers,
        )
        .filter(lambda results: results.ok() and args.version not in results.stdout)
        .hosts()
        .run_task(
            Task(
                "Update app version",
                update_via_apt_command,
                timeout=300,
                retry=5,
                pre_run=log_command_callback,
                post_run=log_status_callback,
            ),
            args.app,
            max_workers=args.max_workers,
        )
        .pipe(lambda results: results2csv(results, args))
    )



"""Viper project APIs
---------------------
This module provides APIs for customizing viper based on any project
requirements. When viper detects a `viperfile.py` file in the current directory,
it scans the file and adds the project specific commands to it's list.

We can see the commands by running `viper --help`. These commands are
generally prefixed with `@` so that they are easily recognizable.

See `viper.demo.viperfile.py` for examples.
"""


from __future__ import annotations
from argparse import ArgumentParser
from argparse import Namespace
from dataclasses import dataclass
from dataclasses import field
from viper.cli_base import SubCommand
from viper.collections import Collection as ViperCollection
from viper.collections import Hosts
from viper.collections import Items

import typing as t


class Arg:
    """Args that can be passed to ArgumentParser"""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


@dataclass
class Project:
    """A project namespace for the sub commands."""

    prefix: str
    hostgroup_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    filter_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    handler_commands: t.List[SubCommand] = field(default_factory=lambda: [])
    job_commands: t.List[SubCommand] = field(default_factory=lambda: [])

    def all_commands(self) -> t.List[SubCommand]:
        """Return all sub commands"""
        return (
            self.hostgroup_commands
            + self.filter_commands
            + self.handler_commands
            + self.job_commands
        )

    def hostgroup(self, args: t.Optional[t.Sequence[Arg]] = None):
        """Use this decorator to define host groups."""

        def wrapper(
            func: t.Callable[[Namespace], Hosts]
        ) -> t.Callable[[Namespace], Hosts]:
            class HostGroupCommand(SubCommand):
                name = f"@{self.prefix}:{func.__name__}"
                __doc__ = f"[@{self.prefix}:{func.__name__} > Hosts] {func.__doc__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg.args, **arg.kwargs)

                def __call__(self, args: Namespace) -> int:
                    print(func(args).to_json())
                    return 0

            self.hostgroup_commands.append(HostGroupCommand)
            return func

        return wrapper

    def filter(self, objtype: t.Type[Items], args: t.Optional[t.Sequence[Arg]] = None):
        """Use this decorator to define filters."""

        if not issubclass(objtype, Items):
            raise ValueError(f"{objtype} does not have filter option")

        def wrapper(
            func: t.Callable[[Items, Namespace], bool],
        ) -> t.Callable[[Items, Namespace], bool]:
            class FilterCommand(SubCommand):
                name = f"@{self.prefix}:{func.__name__}"
                __doc__ = f"[{objtype.__name__} > @{self.prefix}:{func.__name__} > {objtype.__name__}] {func.__doc__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg.args, **arg.kwargs)

                def __call__(self, args: Namespace) -> int:
                    print(
                        objtype.from_json(input())
                        .filter(lambda host: func(host, args))
                        .to_json()
                    )
                    return 0

            self.filter_commands.append(FilterCommand)
            return func

        return wrapper

    def handler(
        self,
        fromtype: ViperCollection,
        totype: t.Optional[type] = None,
        args: t.Optional[t.Sequence[Arg]] = None,
    ):
        """Use this decorator to define handlers."""

        if not issubclass(fromtype, ViperCollection):
            raise ValueError(f"{fromtype} does not have pipe option")

        def wrapper(
            func: t.Callable[[ViperCollection, Namespace], object]
        ) -> t.Callable[[ViperCollection, Namespace], object]:
            class HandlerCommand(SubCommand):
                name = f"@{self.prefix}:{func.__name__}"
                __doc__ = f"[{fromtype.__name__} > @{self.prefix}:{func.__name__} > {totype.__name__}] {func.__doc__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg.args, **arg.kwargs)

                def __call__(self, args: Namespace) -> int:
                    r = fromtype.from_json(input()).pipe(lambda obj: func(obj, args))
                    if totype:
                        print(
                            r.to_json()
                            if issubclass(totype, ViperCollection)
                            else totype(r)
                        )
                    return 0

            self.handler_commands.append(HandlerCommand)
            return func

        return wrapper

    def job(
        self,
        fromtype: t.Type[ViperCollection],
        totype: t.Optional[type] = None,
        args: t.Optional[t.Sequence[Arg]] = None,
    ):
        """Use this decorator to define job."""

        if not issubclass(fromtype, ViperCollection):
            raise ValueError(f"{fromtype} is not a valid input for any job")

        def wrapper(
            func: t.Callable[[ViperCollection, Namespace], object]
        ) -> t.Callable[[ViperCollection, Namespace], object]:
            class WorkFlowCommand(SubCommand):
                name = f"@{self.prefix}:{func.__name__}"
                __doc__ = f"[{fromtype.__name__} > @{self.prefix}:{func.__name__} > {totype.__name__}] {func.__doc__}"

                def add_arguments(self, parser: ArgumentParser) -> None:
                    if args:
                        for arg in args:
                            parser.add_argument(*arg.args, **arg.kwargs)

                def __call__(self, args: Namespace) -> int:
                    if totype:
                        r = func(fromtype.from_json(input()), args)
                        print(
                            r.to_json()
                            if issubclass(totype, ViperCollection)
                            else totype(r)
                        )
                    return 0

            self.job_commands.append(WorkFlowCommand)
            return func

        return wrapper
